#pragma once

#include <cstdlib>
#include <type_traits>
#include <utility>
#include <limits>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include "line.hpp"
#include "../shader.hpp"

namespace BGTU::ComputerGraphicsLabWork::Utilities::Shapes {
	template<class edge_point_receiver_t, class fill_line_receiver_t>
	void iterate_triangle_fill_h_lines(
			RendererApi::PointI<2>::component_t x0, RendererApi::PointI<2>::component_t y0,
			RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t y1,
			RendererApi::PointI<2>::component_t x2, RendererApi::PointI<2>::component_t y2,
			edge_point_receiver_t edge_point_receiver,
			fill_line_receiver_t fill_line_receiver
	) {
		auto _swap = [](RendererApi::PointI<2>::component_t *sx0, RendererApi::PointI<2>::component_t *sy0, RendererApi::PointI<2>::component_t *sx1, RendererApi::PointI<2>::component_t *sy1) {
			RendererApi::PointI<2>::component_t tmp;
			(tmp = *sx0), (*sx0 = *sx1), (*sx1 = tmp);
			(tmp = *sy0), (*sy0 = *sy1), (*sy1 = tmp);
		};

		if (y0 < y1)
			_swap(&x0, &y0, &x1, &y1);
		if (y1 < y2)
			_swap(&x1, &y1, &x2, &y2);
		if (y0 < y1)
			_swap(&x0, &y0, &x1, &y1);
		if (y0 == y1 && x0 < x1 && y0 != y2)
			_swap(&x0, &y0, &x1, &y1);
		if (y1 == y2 && x2 < x1)
			_swap(&x1, &y1, &x2, &y2);


		auto _impl = [&]<bool is_r2l>(
				brezenham_line_iterable::iterator upper_cathetus,
				brezenham_line_iterable::iterator lower_cathetus,
				brezenham_line_iterable::iterator hypotenuse
		) {
			RendererApi::PointI<2>::component_t x, ex, y;

			x = upper_cathetus->x;
			y = upper_cathetus->y;
			edge_point_receiver(x, y);

			UPPER_CATHETUS:
			upper_cathetus++;
			if (upper_cathetus == nullptr) {
				goto LOWER_CATHETUS;
			}
			if (upper_cathetus->y < y) {
				while (true) {
					if constexpr (is_r2l) {
						if (hypotenuse->x < x) {
							edge_point_receiver(hypotenuse->x, hypotenuse->y);
							ex = hypotenuse->x;
							goto HYPOTENUSE_UPPER;
						}
					} else {
						if (hypotenuse->x > x) {
							edge_point_receiver(hypotenuse->x, hypotenuse->y);
							ex = hypotenuse->x;
							goto HYPOTENUSE_UPPER;
						}
					}

					hypotenuse++;
					if (hypotenuse->y < y) {
						edge_point_receiver(upper_cathetus->x, upper_cathetus->y);
						x = upper_cathetus->x;
						y = upper_cathetus->y;
						goto UPPER_CATHETUS;
					}
				}
			}

			edge_point_receiver(upper_cathetus->x, upper_cathetus->y);
			if constexpr (is_r2l) {
				x = ((x < upper_cathetus->x) ? x : (upper_cathetus->x));
			} else {
				x = ((x > upper_cathetus->x) ? x : (upper_cathetus->x));
			}

			goto UPPER_CATHETUS;

			HYPOTENUSE_UPPER:
			hypotenuse++;
			if (hypotenuse->y < y) {
				if constexpr (is_r2l) {
					fill_line_receiver(y, ex + 1, x);
				} else {
					fill_line_receiver(y, x + 1, ex);
				}
				y--;
				x = upper_cathetus->x;
				edge_point_receiver(upper_cathetus->x, y);
				goto UPPER_CATHETUS;
			}

			if constexpr (is_r2l) {
				ex = ((ex > hypotenuse->x) ? ex : (hypotenuse->x));
				if (hypotenuse->x < x)
					edge_point_receiver(hypotenuse->x, hypotenuse->y);
			} else {
				ex = ((ex < hypotenuse->x) ? ex : (hypotenuse->x));
				if (hypotenuse->x > x)
					edge_point_receiver(hypotenuse->x, hypotenuse->y);
			}
			goto HYPOTENUSE_UPPER;

			LOWER_CATHETUS:
			lower_cathetus++;
			if (lower_cathetus == nullptr) {
				ex = x;
				goto HYPOTENUSE_LOWER;
			}
			if (lower_cathetus->y < y) {
				while (true) {
					if constexpr (is_r2l) {
						if (hypotenuse->x < x) {
							edge_point_receiver(hypotenuse->x, hypotenuse->y);
							ex = hypotenuse->x;
							goto HYPOTENUSE_LOWER;
						}
					} else {
						if (hypotenuse->x > x) {
							edge_point_receiver(hypotenuse->x, hypotenuse->y);
							ex = hypotenuse->x;
							goto HYPOTENUSE_LOWER;
						}
					}

					hypotenuse++;
					if (hypotenuse->y < y) {
						edge_point_receiver(lower_cathetus->x, lower_cathetus->y);
						x = lower_cathetus->x;
						y = lower_cathetus->y;
						goto LOWER_CATHETUS;
					}
				}
			}

			edge_point_receiver(lower_cathetus->x, lower_cathetus->y);
			if constexpr (is_r2l) {
				x = ((x < lower_cathetus->x) ? x : (lower_cathetus->x));
			} else {
				x = ((x > lower_cathetus->x) ? x : (lower_cathetus->x));
			}
			goto LOWER_CATHETUS;

			HYPOTENUSE_LOWER:
			hypotenuse++;
			if (hypotenuse == nullptr) {
				return;
			}
			if (hypotenuse->y < y) {
				if constexpr (is_r2l) {
					fill_line_receiver(y, ex + 1, x);
				} else {
					fill_line_receiver(y, x + 1, ex);
				}
				y--;
				x = lower_cathetus->x;
				edge_point_receiver(lower_cathetus->x, y);
				goto LOWER_CATHETUS;
			}
			if constexpr (is_r2l) {
				ex = ((ex > hypotenuse->x) ? ex : (hypotenuse->x));
				if (hypotenuse->x < x)
					edge_point_receiver(hypotenuse->x, hypotenuse->y);
			} else {
				ex = ((ex < hypotenuse->x) ? ex : (hypotenuse->x));
				if (hypotenuse->x > x)
					edge_point_receiver(hypotenuse->x, hypotenuse->y);
			}
			goto HYPOTENUSE_LOWER;
		};

		auto dir = (x0 - x1) * (y2 - y1) - (x2 - x1) * (y0 - y1);
		if (dir > 0) {
			_impl.template operator()<true>(
					brezenham_line_iterable{x0, y0, x1, y1}.begin(),
					brezenham_line_iterable{x1, y1, x2, y2}.begin(),
					brezenham_line_iterable{x0, y0, x2, y2}.begin()
			);
		} else if (dir < 0) {
			_impl.template operator()<false>(
					brezenham_line_iterable{x0, y0, x1, y1}.begin(),
					brezenham_line_iterable{x1, y1, x2, y2}.begin(),
					brezenham_line_iterable{x0, y0, x2, y2}.begin()
			);
		} else {
			iterate_line(x0, y0, x2, y2, edge_point_receiver);
		}
	}

	template<class edge_point_receiver_t, class fill_line_receiver_t>
	void iterate_triangle_fill_h_lines(
			RendererApi::PointI<2> p0,
			RendererApi::PointI<2> p1,
			RendererApi::PointI<2> p2,
			edge_point_receiver_t edge_point_receiver,
			fill_line_receiver_t fill_line_receiver
	) {
		iterate_triangle_fill_h_lines(
				p0.x, p0.y,
				p1.x, p1.y,
				p2.x, p2.y,
				edge_point_receiver,
				fill_line_receiver
		);
	}

	template<class receiver_t>
	void iterate_triangle_fill(
			RendererApi::PointI<2>::component_t x0, RendererApi::PointI<2>::component_t y0,
			RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t y1,
			RendererApi::PointI<2>::component_t x2, RendererApi::PointI<2>::component_t y2,
			receiver_t receiver
	) {
		iterate_triangle_fill_h_lines(
				x0, y0,
				x1, y1,
				x2, y2,
				[&](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { receiver(true, x, y); },
				[&](RendererApi::PointI<2>::component_t y, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t ex) {
					for (; x < ex; x++)
						receiver(false, x, y);
				}
		);
	}

	template<class receiver_t>
	void iterate_triangle_fill(
			RendererApi::PointI<2> p0,
			RendererApi::PointI<2> p1,
			RendererApi::PointI<2> p2,
			receiver_t receiver
	) {
		iterate_triangle_fill(
				p0.x, p0.y,
				p1.x, p1.y,
				p2.x, p2.y,
				receiver
		);
	}

	template<class variable_t>
	struct triangle_barycentric_interpolator {
	private:
		long double full_square;
		RendererApi::PointF<2> p0;
		RendererApi::PointF<2> p1;
		RendererApi::PointF<2> p2;
		variable_t v0;
		variable_t v1;
		variable_t v2;

		struct _empty {
		};

		std::conditional_t<std::is_same_v<variable_t, RendererApi::Color::Transparent>, bool, _empty> skip_alpha;

		static long double calculate_square_2(
				RendererApi::PointF<2> p0,
				RendererApi::PointF<2> p1,
				RendererApi::PointF<2> p2
		) noexcept {
			return (p1.x - p0.x) * (p2.y - p0.y) - (p2.x - p0.x) * (p1.y - p0.y);
		}


	public:
		triangle_barycentric_interpolator(
				RendererApi::PointF<2> p0, RendererApi::PointF<2> p1, RendererApi::PointF<2> p2,
				variable_t v0, variable_t v1, variable_t v2
		) : p0{p0}, p1{p1}, p2{p2}, v0{v0}, v1{v1}, v2{v2}, full_square{calculate_square_2(p0, p1, p2)} {
			if constexpr (std::is_same_v<variable_t, RendererApi::Color::Transparent>) {
				this->skip_alpha = v0.alpha == v1.alpha && v0.alpha == v2.alpha;
			}
		}

		variable_t interpolate_point(RendererApi::PointF<2> p) const {
			const double k0 = calculate_square_2(p, this->p1, this->p2) / this->full_square;
			const double k1 = calculate_square_2(p, this->p2, this->p0) / this->full_square;
			const double k2 = calculate_square_2(p, this->p0, this->p1) / this->full_square;
			auto f = [&]<class intermediate_t>(intermediate_t (variable_t::*offset)) -> intermediate_t {
				return static_cast<intermediate_t>((this->v0.*offset) * k0 + (this->v1.*offset) * k1 + (this->v2.*offset) * k2);
			};
			if constexpr (std::is_same_v<variable_t, RendererApi::Color::Transparent>) {
				RendererApi::Color base{
						f(&RendererApi::Color::Transparent::red),
						f(&RendererApi::Color::Transparent::green),
						f(&RendererApi::Color::Transparent::blue),
				};
				return RendererApi::Color::Transparent{
						base.red, base.green, base.blue,
						(this->skip_alpha ? this->v0.alpha : f(&RendererApi::Color::Transparent::alpha))
				};
			} else if constexpr (std::is_same_v<variable_t, RendererApi::Color>) {
				return RendererApi::Color{
						f(&RendererApi::Color::red),
						f(&RendererApi::Color::green),
						f(&RendererApi::Color::blue)
				};
			} else {
				return static_cast<variable_t>(f(&variable_t::variable_t));
			}
		}

		variable_t interpolate_point(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const {
			return this->interpolate_point(RendererApi::PointF<2>{x, y});
		}
	};

	class triangle_barycentric_shader : public Shader {
	private:
		triangle_barycentric_interpolator<RendererApi::Color::Transparent> interpolator;
	public:
		inline triangle_barycentric_shader(
				RendererApi::PointF<2> p0, RendererApi::PointF<2> p1, RendererApi::PointF<2> p2,
				RendererApi::Color::Transparent c0, RendererApi::Color::Transparent c1, RendererApi::Color::Transparent c2
		) noexcept: interpolator{p0, p1, p2, c0, c1, c2} {}


		[[nodiscard]] inline RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const final {
			return this->interpolator.interpolate_point(x, y);
		}
	};
}